package ec3.client;

import ec3.shared.*;

import java.util.Vector;

/**
 * Das EasyChat-Protokoll
 * 
 * @since 0.2
 * @version 0.6
 */
public class ClientImpl implements EC3Listener, ImplCallBack {
	
	/**
	 * Versionsstring des Protokolls
	 */
	public static final String VERSION = "0.6 fr Java";
	
	/**
	 * interne Versionsnummernliste des Protokolls
	 */
	public static final String PROTO_VERSION = "0.6,0";
	
	/**
	 * Standardport der EasyChat-Server
	 */
	public static final int DEFAULT_PORT = 1000;
	
	/**
	 * Konstruktor
	 *
	 * @param e Referenz auf das FrontEnd-Objekt
	 * @param protocolExtensions String-Array, das die Klassennamen aller zu ladenden ProtocolExtensions beinhaltet
	 */
	public ClientImpl (EC3Listener e, String[] protocolExtensions) {
		
		//EventDealer bauen
		dealer = new EC3EventDealer ();
		dealer.addEC3Listener (e);
		
		//EC3Socket erstellen
		socket = new EC3Socket ();
		socket.addEC3Listener (this);
		
		//Eigenschaftsobjekt initalisieren
		clientProperties = new ec3.shared.Properties ();
		clientProperties.setProperty ("client.connected", "0");
		clientProperties.setProperty ("client.loggedIn", "0");
		clientProperties.setProperty ("client.inChatRoom", "0");
		
		//Public-Konstanten im Eigenschaftsobjekt sichern
		clientProperties.setProperty ("client.supportedProtoVersion", PROTO_VERSION);
		clientProperties.setProperty ("client.version", VERSION);
		clientProperties.setProperty ("client.defaultPort", String.valueOf (DEFAULT_PORT));
		
		//Liste der Protokollerweiterungen bauen
		this.protocolExtensions = new Vector ();
		Class protocolExtension = ProtocolExtension.class;
		for (int i = 0; i < protocolExtensions.length; i++) {
			
			try {
				
				if (protocolExtensions[i] != null) {
					
					Class clazz = Class.forName (protocolExtensions[i]);
					if (protocolExtension.isAssignableFrom (clazz)) {
						
						ProtocolExtension ex = (ProtocolExtension)clazz.newInstance ();
						this.protocolExtensions.addElement (ex);
//						System.out.println ("ProtocolExtension " + protocolExtensions[i] + " geladen");
						
					} else {
						
						System.out.println ("ProtocolExtension " + protocolExtensions[i] + " nicht geladen (!isAssignableFrom).");
						
					}
					
				}
				
			} catch (ClassNotFoundException exception) {
				
				System.out.println ("ProtocolExtension " + protocolExtensions[i] + " nicht geladen (ClassNotFoundException).");
				
			} catch (InstantiationException exception) {
				
				System.out.println ("ProtocolExtension " + protocolExtensions[i] + " nicht geladen (InstantiationException).");
				
			} catch (IllegalAccessException exception) {
				
				System.out.println ("ProtocolExtension " + protocolExtensions[i] + " nicht geladen (IllegalAccessException).");
				
			}
			
		}
		
	}//ClientImpl ()
	
	/**
	 * erstellt eine Verbindung zum angegebenen Server
	 * 
	 * @param server der Server, zu dem verbunden werden soll
	 * @param port der Port, auf dem der Server luft (1-65536 oder 0 = Standard)
	 */
	public void connect (String server, int port) {
		
		if ((port < 0) || (port > 65536)) {
			
			throw new IllegalArgumentException ("Port not in specified Range");
			
		}
		if (port == 0) {
			
			port = DEFAULT_PORT;
			
		}
		
		//ew. vorher alte Verbindung trennen
		if (clientProperties.getProperty ("client.connected", "0")
		 .equals ("1")) {
			
			disconnect ();
			
		}//if
		
		//Eigenschaften setzen
		clientProperties.setProperty ("client.connected", "1");
		clientProperties.setProperty ("client.loggedIn", "0");
		clientProperties.setProperty ("client.inChatRoom", "0");
		clientProperties.setProperty ("client.ignoreDisconnect", "0");
		
		//Event senden
		dealer.fireEC3Event (this, "connect", new String [] {});
		
		//verbinden
		socket.open (server, port);
		
		//Eigenschaften setzen
		clientProperties.setProperty ("server.adress", server);
		clientProperties.setProperty ("server.port", new Integer (port)
		 .toString ());
		
		//ProtocolExtensions benachrichtigen
		for (int i = 0; i < protocolExtensions.size (); i++) {
			
			((ProtocolExtension)(protocolExtensions.elementAt (i)))
			 .connect (clientProperties, this);
			
		}
		
	}//connect (...)
	
	/**
	 * trennt eine Verbindung
	 */
	public void disconnect () {
		
		//nur wenn eine Verbindung besteht
		if (clientProperties.getProperty ("client.connected", "0")
		 .equals ("1")) {
			
			//Eigenschaften setzen
			clientProperties.setProperty ("client.connected", "0");
			clientProperties.setProperty ("client.loggedIn", "0");
			
			//Socket schlieen
			socket.close ();
			
			//Event absetzen
			dealer.fireEC3Event (this, "disconnect", new String [] {});
			
			//ProtocolExtensions benachrichtigen
			for (int i = 0; i < protocolExtensions.size (); i++) {
				
				((ProtocolExtension)(protocolExtensions.elementAt (i)))
				 .disconnect (clientProperties, this);
				
			}
		
		}//if
		
	}//disconnect ()
	
	/**
	 * setzt Servername/port und bestimmt das Vorhandensein eines Chatnamens vor dem 1. verbinden
	 *
	 * @param server der Server
	 * @param port der Port des Servers
	 * @param hasChatName ist schon ein Chatname im System?
	 */
	public void setInitData (String server, int port, boolean hasChatName) {
		
		//Eigenschaften setzen
		clientProperties.setProperty ("server.adress", server);
		clientProperties.setProperty ("server.port", new Integer (port)
		 .toString ());
		clientProperties.setProperty ("client.hasChatName", hasChatName ? "1" : "0");
	}
	
	/*
	 * ermglicht den protocolExtension's den Zugriff auf das Socket-Objekt
	 *
	 * @param command zu sendender Befehl
	 * @param data zu sendende Daten
	 */
	public void sockCommand (String command, String data) {
		
		socket.send (command, data);
		
	}
	
	/*
	 * ermglicht den protocolExtension's den Zugriff auf das Socket-Objekt
	 *
	 * @param command zu sendender Befehl
	 * @param data zu sendende Daten
	 */
	public void sockCommand (String command, byte[] data) {
		
		socket.send (command, data);
		
	}
	
	/*
	 * ermglicht den protocolExtension's den Zugriff auf das frontEnd-Objekt
	 *
	 * @param command zu sendender Befehl
	 * @param data zu sendende Daten
	 */
	public void frontEndCommand (String command, String[] data) {
		
		dealer.fireEC3Event (this, command, data);
		
	}
	
	/* (non-Javadoc)
	 * @see ec3.client.ImplCallBack#pluginCommand(java.lang.String, java.lang.String[])
	 */
	public boolean pluginCommand(String command, String[] data) {
		
		// ProtocolExtensions nacheinander benachrichtigen
		boolean done = false;
		for (int i = 0; i < protocolExtensions.size () && !done; i++) {
			done = ((ProtocolExtension)protocolExtensions.elementAt(i))
			.pluginCommand(
				command,
				data,
				clientProperties,
				this
			);
		}
		
		return done;
		
	}
	
	/* (non-Javadoc)
	 * @see ec3.client.ImplCallBack#enableEncryption(ec3.shared.CryptoHook)
	 */
	public void enableEncryption(CryptoHook hook) {
		this.socket.setCryptoHook(hook);
	}
	
	/**
	 * liest eine Eigenschaft aus
	 *
	 * @param property der Name der Eigenschaft
	 * @return die Eigenschaft oder null, wenn die Eigenschaft nicht gesetzt ist
	 */
	public String getProperty (String property) {
		
		return clientProperties.getProperty (property);
		
	}
	
	/**
	 * liest eine Eigenschaft aus
	 * gibt default zurck, wenn die Eigenschaft nicht vorhanden ist
	 *
	 * @param property der Name der Eigenschaft
	 * @param defaultValue der Standardwert, wenn die Eigenschaft nicht
	 *        gesetzt ist
	 * @return die Eigenschaft oder der Standardwert, wenn die Eingenschaft
	 *         nicht gesetzt ist
	 */
	public String getProperty (String property, String defaultValue) {
		
		return clientProperties.getProperty (property, defaultValue);
		
	}
	
	/**
	 * gibt das interne Properties-Objekt zurck
	 * 
	 * @return das Properties-Objekt
	 */
	public Properties getProperties () {
		
		return clientProperties;
		
	}
	
	public void EC3CallBack (EC3Event e) {
		
		//Event auslesen
		Object source = e.getSource ();
		String event = e.getEvent ();
		String[] data = e.getData ();
		
		//Untersuchen, von wo das Event kam
		//Socket
		if (source == socket) {
			
			//Socket geschlossen
			if (event.equals ("closed")) {
				
				//trennen
				disconnect ();
				
			//Socket mit Fehler geschlossen
			} else if (event.equals ("aborted")) {
				
				String[] dataSplit = Tools.split (data[0], Tools.CHR0);
				if (!dataSplit[0].equals("serverClosedConnection")
				 || clientProperties.getProperty ("client.ignoreDisconnect", "0").equals("0")) {
					
					//Fehlerevent senden
					dealer.fireEC3Event (this, "aborted", dataSplit);
					
				}
				//trennen
				disconnect ();
				
				// ProtocolExtensions benachrichtigen
				for (int i = 0; i < protocolExtensions.size (); i++) {
					((ProtocolExtension)protocolExtensions.elementAt(i))
					.sockError(
						data[0],
						clientProperties,
						this
					);
					
				}
				
			//Socket hat Nachricht empfangen
			} else if (event.equals ("recieved")) {
				
				boolean processed = false;
				
				//ProtocolExtension suchen
				for (
					int i = 0;
					(i < protocolExtensions.size ()) && (!processed);
					i++
				) {
					
					processed = ((ProtocolExtension)
						(protocolExtensions.elementAt (i))
					).sockCommand (
						data[0],
						data[1],
						clientProperties,
						this
					);
					
				}
				
				//wenn keine gefunden => Meldung
				if (!processed) {
					
					//Meldung absetzen
					dealer.fireEC3Event (
						this,
						"unknownCommandFromSocket",
						data
					);
					
				}
				
			} else {
				System.err.println("Kann net sein: " + event); // DEBUG
			}//if event ==
			
		//source != socket => Frontend
		} else {
			
			boolean processed = false;
			
			for (
				int i = 0;
				(i < protocolExtensions.size ()) && (!processed);
				i++
			) {
				
				processed = ((ProtocolExtension)
					(protocolExtensions.elementAt (i))
				).frontEndCommand (
					event,
					data,
					clientProperties,
					this
				);
				
			}
			if (!processed) {
				
				String[] temp = new String[data.length + 1];
				temp[0] = event;
				System.arraycopy (data, 0, temp, 1, data.length);
				//Meldung absetzen
				dealer.fireEC3Event (
					this,
					"unknownCommandFromFrontEnd",
					temp
				);
				
			}
			
		}//source ==
		
	}//EC3CallBack (...)
	
	/**
	 * der Verteiler fr Events
	 */
	private EC3EventDealer dealer;
	
	/**
	 * das Socket-Objekt
	 */
	private EC3Socket socket;
	
	/**
	 * die Liste aller internen Eigenschaften
	 */
	private Properties clientProperties;
	
	/**
	 * die Liste aller ProtocolExtensions
	 */
	private Vector protocolExtensions;

	/* (non-Javadoc)
	 * @see ec3.client.ImplCallBack#connect()
	 */
	public void connect() {
		String server = clientProperties.getProperty("server.adress");
		String portStr = clientProperties.getProperty("server.port");
		if (server == null || portStr == null) {
			throw new IllegalStateException("Kein Server zum Verbinden bekannt");
		}
		int port = Integer.parseInt(portStr);
		connect(server, port);
	}
	
}//ClientImpl