package ec3.shared;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.Socket;

/**
 * Kapselung der Socketkommunikation und des primren Protokolls
 *
 * @since 0.1
 */
public final class EC3Socket {
	
	/**
	 * Der zu benutzende Zeichensatz fr die Socket-Kommunikation. Sollte ein
	 * Ein-Byte-Zeichensatz sein.
	 */
	public static final String CHARSET = "windows-1252";
	
	/**
	 * Das Verschlsselungsobjekt.
	 */
	CryptoHook cryptoHook;
	
	/**
	 * erstellt das Socket-Objekt, ohne die Verbindung zu ffnen
	 */
	public EC3Socket () {
		super ();
		//Event
		dealer = new EC3EventDealer ();
		//Socket: geschlosssen = nicht vorhanden
		socket = null;
		reader = null;
		writer = null;
		opened = false;
		opening = false;
		//Lesethread
		readThread = null;
		// die Default-Implementation nimmt keine Verschlsselung vor
		cryptoHook = new CryptoHook() {
			/* (non-Javadoc)
			 * @see ec3.shared.CryptoHook#encrypt(byte[])
			 */
			public byte[] encrypt(byte[] data) {
				return data;
			}
			/* (non-Javadoc)
			 * @see ec3.shared.CryptoHook#decrypt(byte[])
			 */
			public byte[] decrypt(byte[] data) {
				return data;
			}
		};
		
		try {
			String[] cacheProps = new String[] {
				"sun.net.inetaddr.ttl", // 1.3 + 1.4
				"sun.net.inetaddr.negative.ttl", // 1.3 + 1.4
				"networkaddress.cache.ttl", // 1.4 + 1.5
				"networkaddress.cache.negative.ttl" // 1.4 + 1.5
			};
			for (int i = 0; i < cacheProps.length; i++) {
				System.setProperty(cacheProps[i], "0"); // 0 = unbegrenzt
			}
		} catch (Throwable t) {
			t.printStackTrace();
		}
	}//EC3Socket ()
	
	/**
	 * erstellt das Socket-Objekt und ffnet die Verbindung
	 * 
	 * @param host der Zielhost
	 * @param port der ZIelport
	 */
	public EC3Socket (String host, int port) {
		this();
		//Socket: ffnen
		open(host, port);
	}//EC3Socket (String host, int port)
	
	/**
	 * erstellt das Socket-Objekt unter Verwendung des gegebenen Sockets
	 * 
	 * @param sock das zu benutzende Socket
	 */
	public EC3Socket (Socket sock) {
		this();
		//Socket: schon offen
		open(sock);
	}//EC3Socket (Socket sock)
	
	
	/**
	 * ffnet die Verbindung
	 * 
	 * @param host der Zielhost
	 * @param port der Zielport
	 */
	public void open (String host, int port) {
		//ew. alte Verbindung schlieen
		if (opened || opening) {close ();}
		//neue Verbindung ffnen
		try {
			opened = false;
			opening = true;
			//Lesethread starten
			readThread = new ReadThread (host, port, this, dealer);
			readThread.start ();
		} catch (SecurityException e) {
			abort ("security" + Tools.CHR0 + host + ":" + port);
		}//try
	}//open (String host, int port)
	
	/**
	 * benutzt schon vorhandenes Socket
	 * 
	 * @param sock das zu benutzende Socket
	 */
	public void open (Socket sock) {
		//ew. alte Verbindung schlieen
		if (opened || opening) {close ();}
		//neue Verbindung ffnen
		try {
			opened = true;
			opening = false;
			socket = sock;
			socket.setSoTimeout (1000);
			reader = new BufferedInputStream (socket.getInputStream ());
			writer = new BufferedOutputStream (socket.getOutputStream ());
			opened = true;
			//Lesethread starten
			readThread = new ReadThread (reader, this, dealer);
			readThread.start ();
		} catch (IOException e) {
			abort ("IOException in public void open (Socket sock): "
				+ e.toString ());
		}//try
	}//open (Socket sock)
	
	/**
	 * schliet die Verbindung
	 */
	public void close () {
		if (opened || opening) {
			//ReadThread beenden
			if (readThread != null) {
				readThread.interrupt ();
			}
			readThread = null;
			//Socket, Reader, Writer schlieen
			try {
				if (reader != null) {
					reader.close ();
				}//if
			} catch (IOException e) {
				// ignoriert
			}
			reader = null;
			try {
				if (writer != null) {
					writer.close ();
				}//if
			} catch (IOException e) {
				// ignoriert
			}
			writer = null;
			try {
				if (socket != null) {
					socket.close ();
				}//if
			} catch (IOException e) {
				// ignoriert
			}
			socket = null;
			opened = false;
			opening = false;
			dealer.fireEC3Event (this, "closed", null);
		}//if
	}//close ()
	
	/**
	 * formatiert und sendet eine Nachricht
	 * 
	 * @param command der Befehl
	 * @param data die Daten zum Befehl
	 */
	public void send (String command, String data) {
		/*
		//nur, wenn das Socket existiert
		if (opened && (socket != null) && (writer != null)) {
			writer.print (command + Tools.CRLF + data.length () + Tools.CRLF
				+ Tools.CHR2 + Tools.CRLF + data);
			writer.flush ();
			//Fehler aufgetreten?
			if (writer.checkError ()) {
				abort ("sendError");
			}
		}//if
		*/
		try {
			send(command, data.getBytes(CHARSET));
		} catch (UnsupportedEncodingException e) {
			// ignoriert
		}
	}//send (String command, String data)
	
	/**
	 * formatiert und sendet Byte-Daten
	 * 
	 * @param command der Befehl
	 * @param data die Daten zum Befehl
	 */
	public void send (String command, byte[] data) {
		//nur, wenn das Socket existiert
		if (opened && (socket != null) && (writer != null)) {
			data = cryptoHook.encrypt(data);
			
			String header = command + Tools.CRLF + data.length + Tools.CRLF + Tools.CHR2
			 + Tools.CRLF;
			
			try {
				writer.write(header.getBytes(CHARSET));
				writer.write(data);
				writer.flush();
			} catch (IOException e) {
				//Fehler aufgetreten?
				abort ("sendError");
			}
//			System.out.println("Gesendet: " + Tools.escape(header + new String(data))); // DEBUG
		}//if
	}//send (String command, byte[] data)
	
	/**
	 * Setzt den zu benutzenden CryptoHook.
	 * 
	 * @param hook der CryptoHook, der benutzt werden soll-
	 */
	public void setCryptoHook(CryptoHook hook) {
		this.cryptoHook = hook;
	}
	
	/**
	 * fgt einen EventListener hinzu
	 * 
	 * @param e der neue Listener
	 */
	public void addEC3Listener (EC3Listener e) {
		dealer.addEC3Listener (e);
	}//addEC3Listener (...)
	
	/**
	 * entfernt einen EventListener
	 * 
	 * @param e der zu entfernende Listener
	 */
	public void removeEC3Listener (EC3Listener e) {
		dealer.removeEC3Listener (e);
	}//removeEC3Listener (...)
	
	/* 
	 * (non-Javadoc)
	 * @see java.lang.Object#finalize()
	 */
	protected void finalize () throws Throwable {
		super.finalize ();
		//beim beenden Socket schlieen!!
		socket.close ();
	}//finalize ()
	
	/**
	 * schliet die Verbindung mit einer Fehlermeldung
	 * 
	 * @param reason die Fehlermesdung mit Hinweis, warum die Verbindung beendet
	 *        wurde
	 */
	void abort (String reason) {
		
		//ReadThread beenden
		if ((readThread != null) && (readThread.isAlive())) {
			
			readThread.interrupt ();
			readThread = null;
			
		}
		
		//Socket, Reader, Writer schlieen
		if (reader != null) {
			
			try {
				
				reader.close ();
				
			} catch (IOException e) {
				// ignoriert
			}
			reader = null;
			
		}
		
		if (writer != null) {
			
			try {
				writer.close ();
			} catch (IOException e) {
				// ignoriert
			}
			writer = null;
			
		}
		
		if (socket != null) {
			
			try {
				
				socket.close ();
				
			} catch (IOException e) {
				// ignoriert
			}
			socket = null;
			
		}
		
		opened = false;
		opening = false;
		dealer.fireEC3Event (this, "aborted", new String[] {reason});
		
	}//abort (...)
	
	/**
	 * wird vom ReadThread aufgerufen, wenn dieser das Socket erfolgreich
	 * geffnet hat
	 * 
	 * @param sock das geffnete Socket
	 * @param r der vom Thread erzeugte Reader (zum Schlieen)
	 */
	void socketOpened(Socket sock, BufferedInputStream r) {
		socket = sock;
		reader = r;
		try {
			writer = new BufferedOutputStream (sock.getOutputStream ());
			opened = true;
			opening = false;
		} catch (IOException e) {
			abort ("IOException in EC3Socket.socketOpened(...): "
				+ e.toString ());
		}
	}
	
	/**
	 * das benutzte Socket
	 */
	private Socket socket;
	
	/**
	 * der Reader aus dem (per ReadThread) gelesen wird
	 */
	private BufferedInputStream reader;
	
	/**
	 * der Writer, in den geschrieben wird
	 */
	private BufferedOutputStream writer;
	
	/**
	 * Lesethread
	 */
	private Thread readThread;
	
	/**
	 * Eventbehandlung
	 */
	private EC3EventDealer dealer;
	
	/**
	 * Flag, ob das Socket offen ist
	 */
	private boolean opened;
	
	/**
	 * Flag, ob das Socket gerade geffnet wird
	 */
	private boolean opening;
	
}//EC3Socket


