////#INCLUDE ../../../defines
package ec3.shared;

import java.io.*;
import java.net.*;

import ec4.tools.version.VersionController;

/**
 * Thread, der periodisch das Socket abfragt
 *
 * @since 0.1
 */
class ReadThread extends Thread {
	
	/**
	 * die maximale Lnge eines Befehls; wird der Befehl lnger, wird das
	 * komplette Paket verworfen
	 */
	private static final int MAX_COMMAND_LENGTH = 128;
	
	/**
	 * die maximale Lnge der Daten; sind die Daten lnger, wird das Paket
	 * verworfen
	 */
	private static final int MAX_DATA_LENGTH = 128*1024; // 128 kiB
	
	/**
	 * die maximale Lnge der Lngenangabe fr die Daten; ist die Angabe lnger,
	 * wird das Paket verworfen
	 */
	private static final int MAX_DATALENGTH_LENGTH = 7;
	
	/**
	 * der aktuelle Lesepuffer
	 */
	private byte[] buffer;
	
	/**
	 * Die Zahl der Bytes im Puffer.
	 */
	private int bufferSize;
	
	/**
	 * Eventbehandlung
	 */
	private EC3EventDealer dealer;
	
	/**
	 * das aktuell gelesene Paket
	 */
	private EC3Packet packet;
	
	/**
	 * das aufrufende Objekt
	 */
	private EC3Socket parent;
	
	/**
	 * der @link Reader aus dem gelesen werden soll
	 */
	private BufferedInputStream reader;
	
	/**
	 * das Socket, welches geffnet werden soll
	 */
	private Socket sock;
	
	/**
	 * der Host, zu dem verbunden werden soll
	 */
	private String host;
	
	/**
	 * der Port, zu dem verbunden werden soll
	 */
	private int port;
	
	/**
	 * Konstruktor fr vorhandenes Socket
	 *
	 * @param reader der {@link java.io.Reader} aus dem gelesen werden soll
	 * @param parent das Objekt, in dessen Namen Botschaften gesendet werden
	 * @param dealer der Event-Verteiler, der benachrichtigt werden soll
	 */
	public ReadThread (BufferedInputStream reader, EC3Socket parent,
	 EC3EventDealer dealer) {
		super ();
		this.reader = reader;
		this.parent = parent;
		this.dealer = dealer;
		sock = null;
		host = null;
		port = -1;
		buffer = new byte[65536];
		bufferSize = 0;
		packet = new EC3Packet ();
		try {
			setPriority (MIN_PRIORITY);
		} catch (IllegalArgumentException e) {
			// ignoriert
		} catch (SecurityException  e) {
			// ignoriert
		}//try
	}//ReadThread ()
	
	/**
	 * Konstruktor fr neues Socket
	 * 
	 * @param host der Host, zu dem verbunden werden soll
	 * @param port der Port, zu dem verbunden werden soll
	 * @param parent das Objekt, in dessen Namen Botschaften gesendet werden
	 * @param dealer der Event-Verteiler, der benachrichtigt werden soll
	 */
	public ReadThread (String host, int port, EC3Socket parent, EC3EventDealer dealer) {
		super ();
		reader = null;
		this.parent = parent;
		this.dealer = dealer;
		sock = null;
		this.host = host;
		this.port = port;
		buffer = new byte[65536];
		bufferSize = 0;
		packet = new EC3Packet ();
		try {
			setPriority (MIN_PRIORITY);
		} catch (IllegalArgumentException e) {
			// ignoriert
		} catch (SecurityException  e) {
			// ignoriert
		}//try
	}//ReadThread ()
	
	public void run () {
		// Netzwerk-Berechtigung holen
		VersionController.getInstance().assertNetworkPerm();
		if (reader == null) {
			try {
				// Host in IP auflsen
				InetAddress a = InetAddress.getByName(host);
				dealer.fireEC3Event(parent, "recieved", new String[] {"__lookup", a.toString()});
				// Verbinden
				sock = new Socket(a, port);
				sock.setSoTimeout (1000);
				reader = new BufferedInputStream(sock.getInputStream());
			} catch (ConnectException e) {
				if (isInterrupted()) return;
				parent.abort ("connectionRefused" + Tools.CHR0 + host + ":" + port);
				return;
			} catch (SocketException e) {
				if (isInterrupted()) return;
				parent.abort ("networkUnreachable" + Tools.CHR0 + host);
				return;
			} catch (UnknownHostException e) {
				if (isInterrupted()) return;
				parent.abort ("unknownHost" + Tools.CHR0 + host);
				return;
			} catch (IOException e) {
				if (isInterrupted()) return;
				parent.abort ("IOException in ReadThread.run(): "
					+ e.toString ());
				return;
			} catch (SecurityException e) {
				if (isInterrupted()) return;
				parent.abort ("security" + Tools.CHR0 + host);
				return;
			}//try
			parent.socketOpened(sock, reader);
		}//if
		//Endlosschleife starten
		byte[] buf = new byte[65536];
		while (!isInterrupted()) {
			try {
				int read = reader.read (buf);
				//Socket wurde geschlossen
				if (read == -1) {
					parent.abort ("serverClosedConnection");
					break;
				//Daten gelesen
				}
//				System.out.println("Empfangen: " + Tools.escape(new String(buf, 0, read))); //DEBUG
				parseBuffer (buf, read);
			} catch (InterruptedIOException e) {
				// ignoriert
			} catch (SocketException e) {
				parent.abort ("SocketException" + Tools.CHR0 + e.getMessage ());
				break;
			} catch (IOException e) {
				parent.abort ("IOException in ReadThread.run (): "+e.toString ());
				break;
			}//try
		}//while
	}//run ()
	
	/**
	 * parst den aktuellen Puffer entsprechend dem Protokoll und sendet eine
	 * Nachricht, wenn ein Paket komplett ist
	 * 
	 * Annahme whrend des Parsens: EC3Socket.CHARSET ist ein Ein-Byte-Zeichensatz
	 * 
	 * @param buf die gelesenen Bytes
	 * @param read die Anzahl der gelesenen Bytes.
	 */
	private void parseBuffer (byte[] buf, int read) {
		int pos = 0; // die aktuelle Leseposition im Puffer
		
		if (bufferSize + read < buffer.length) {
			System.arraycopy(buf, 0, buffer, bufferSize, read);
		} else {
			byte[] temp = new byte[read + bufferSize];
			System.arraycopy(buffer, 0, temp, 0, bufferSize);
			System.arraycopy(buf, 0, temp, bufferSize, read);
			buffer = temp;
		}
		bufferSize += read;
		
		while (bufferSize > pos) {
			String s = "";
			try {
				s = new String(buffer, EC3Socket.CHARSET);
			} catch (UnsupportedEncodingException e1) {
				// ignoriert
			}
			//Zeilenumbruch suchen
			int test = s.indexOf (Tools.CRLF, pos);
			//Befehl erwartet?
			if (packet.command == null) {
				//Ende des Befehls gefunden
				if (test >= 0) {
					//Test auf zu lange Befehlsnamen
					if (test - pos > MAX_COMMAND_LENGTH) {
						parent.abort ("maxCommandLength");
						break;
					}
					//Befehl speichern
					packet.command = s.substring (pos, test);
//					System.out.println("command: " + packet.command); // DEBUG
					pos = test + 2;
					test = s.indexOf (Tools.CRLF, pos);
				//Befehl ist zu lang?	
				} else if (bufferSize - pos > MAX_COMMAND_LENGTH + 2) {
					parent.abort ("maxCommandLength");
					break;
				//sonst: warten
				} else {
					break;
				}
			}
			//Datenlnge erwartet?
			if (packet.dataLength == -1) {
				//Ende der Lnge gefunden gefunden
				if (test >= 0) {
					try {
						//Lnge speichern
						packet.dataLength = Integer.parseInt (
							s.substring (pos, test)
						);
//						System.out.println("datenlnge: " + packet.dataLength);//DEBUG
						packet.data = new byte[packet.dataLength];
						packet.dataDone = 0;
						pos = test + 2;
						test = s.indexOf (Tools.CRLF, pos);
						//max. Lnge berprfen
						if (packet.dataLength > MAX_DATA_LENGTH + 2) {
							parent.abort ("maxDataLength");
							break;
						}
					} catch (NumberFormatException e) {
						parent.abort ("badDataLength");
						break;
					}
				//Lnge ist zu lang?	
				} else if (bufferSize - pos > MAX_DATALENGTH_LENGTH + 2) {
					parent.abort ("badDataLength");
					break;
				//ansonsten: warten
				} else {
					break;
				}
			}
			//Headerendezeichen erwartet?
			if (packet.headerEndMark == false) {
				if (bufferSize - pos >= 3) {
					if (test == pos + 1 && s.charAt(pos) == Tools.CHR2.charAt(0)) {
						packet.headerEndMark = true;
						pos += 3;
//						System.out.println("headertrennzeichen");//DEBUG
					} else {
						parent.abort ("noHeaderEndMark");
						break;
					}
				//sonst: warten
				} else {
					break;
				}
			}
			test = packet.dataLength - packet.dataDone;
			//noch Daten zu lesen?
			if (packet.headerEndMark && (test > 0)) {
				if (bufferSize - pos > test) {
					System.arraycopy(buffer, pos, packet.data, packet.dataDone, test);
//					System.out.println("daten: " + test);//DEBUG
					pos += test;
					packet.dataDone += test;
				} else {
					System.arraycopy(buffer, pos, packet.data, packet.dataDone, bufferSize - pos);
//					System.out.println("daten: " + (bufferSize - pos));//DEBUG
					packet.dataDone += bufferSize - pos;
					bufferSize = 0;
					pos = 0;
				}
			}
			//Paket vollstndig?
			if (packet.headerEndMark
			 && (packet.dataLength == packet.dataDone)) {
				packet.data = parent.cryptoHook.decrypt(packet.data);
				String data = "";
				try {
					data = new String(packet.data, EC3Socket.CHARSET);
				} catch (UnsupportedEncodingException e) {
					// ignoriert
				}
				dealer.fireEC3Event (
					parent,
					"recieved",
					new String[] {packet.command, data}
				);
				//neues Paket anfangen
				packet = new EC3Packet ();
			}
		}
		if (bufferSize == pos) {
			bufferSize = 0;
		} else if (pos > 0) {
			System.arraycopy(buffer, pos, buffer, 0, bufferSize - pos);
			bufferSize -= pos;
		}
		pos = 0;
	}
	
}//ReadThread
	