package ec3.shared.crypto;

/**
 * bietet Base64- und Hex-Codierung
 *
 * @since 0.4
 */
public class B64Hex {
	
	/**
	 * Konstante fr Base64
	 */
	public static final String B64 = "b64";
	
	/**
	 * Konstante fr Hex
	 */
	public static final String HEX = "hex";
	
	/**
	 * Konstante fr "kein Encoding"
	 */
	public static final String PLAIN = "plain";
	
	/**
	 * Zeichenliste fr Base64-Codierung
	 */
	private static final byte[] B64TABLE
	//  0   4   8   C   0   4   8   C   0   4   8   C   0   4   8   C   =
	 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
	 .getBytes();
	
	/**
	 * Zeichenliste fr HEX-Codierung
	 */
	private static final byte[] HEXTABLE = "0123456789ABCDEF".getBytes();
	
	/**
	 * codiert ein byte-Array endsprechend der Angabe in einen String
	 * 
	 * @param s_bytes die zu codierenden Daten
	 * @param how die Codierung (B64, HEX, PLAIN)
	 * @return die codierten Daten als String
	 * @exception IllegalArgumentException wenn die Codierung nicht untersttzt
	 *            wird
	 */
	public static String encode (byte[] s_bytes, String how) {
		if (how.toLowerCase ().equals ("hex")) {
			return hexEncode (s_bytes);
		} else if (how.toLowerCase ().equals ("b64")) {
			return b64Encode (s_bytes);
		} else if (how.toLowerCase ().equals ("plain")) {
			return plainEncode (s_bytes);
		} else {
			throw (new IllegalArgumentException (
				how + "-encoding not supported"
			));
		}
	}//encode (...)
	
	/**
	 * decodiert einen String endsprechend der Angabe in ein byte-Array
	 * 
	 * @param s der zu decodierende String
	 * @param how die Codierung (B64, HEX, PLAIN)
	 * @return die decodierten Daten als byte-Array
	 * @exception IllegalArgumentException wenn die Codierung nicht untersttzt
	 *            wird, der String ungltige Zeichen enthlt oder eine ungltige
	 *            Lnge hat
	 */
	public static byte[] decode (String s, String how) {
		if (how.equals ("hex")) {
			return hexDecode (s);
		} else if (how.equals ("b64")) {
			return b64Decode (s);
		} else if (how.equals ("plain")) {
			return plainDecode (s);
		} else {
			throw (new IllegalArgumentException (
				how + "-decoding not supported"
			));
		}
	}//decode (...)
	
	/**
	 * codiert ein byte-Array in einen Hex-String
	 * 
	 * @param s_bytes die zu codierenden Daten
	 * @return die codierten Daten als String
	 */
	public static String hexEncode (byte[] s_bytes) {
		byte[] result = new byte[s_bytes.length * 2];
		for (int i = 0; i < s_bytes.length; i++) {
			result[i * 2] = HEXTABLE[(s_bytes[i] & 0xF0) >> 4];
			result[i * 2 + 1] = HEXTABLE[s_bytes[i] & 0x0F];
		}
		return new String (result);
	}//hexEncodde (...)
	
	/**
	 * decodiert einen Hex-String in ein byte-Array
	 * 
	 * @param s der zu decodierende String
	 * @return die decodierten Daten als byte-Array
	 * @exception IllegalArgumentException wenn der String ungltige Zeichen
	 *            enthlt oder eine ungltige Lnge hat
	 */
	public static byte[] hexDecode (String s) {
		if ((s.length () % 2) != 0) {
			throw (new IllegalArgumentException (
				"hexDecode: illegal String format"
			));
		}
		byte[] s_bytes = s.getBytes ();
		byte[] result = new byte[s.length () / 2];
		byte[] buf = new byte[2];
		for (int i = 0; i < (s.length () / 2); i++) {
			System.arraycopy (s_bytes, 2 * i, buf, 0, 2);
			for (int j = 0; j < 2; j++) {
				if ((buf[j] > 47) && (buf[j] < 58)) {
					buf[j] -= 48;
				} else if ((buf[j] > 64) && (buf[j] < 71)) {
					buf[j] -= 55;
				} else if ((buf[j] > 96) && (buf[j] < 103)) {
					buf[j] -= 87;
				} else {
					throw (new IllegalArgumentException (
						"hexDecode: illegal String format"
					));
				}
			}
			result[i] = (byte)((buf[0] << 4) | buf[1]);
		}
		return result;
	}//hexDecode (...)
	
	/**
	 * codiert ein byte-Array in einen Base64-String
	 * 
	 * @param s_bytes die zu codierenden Daten
	 * @return die codierten Daten als String
	 */
	public static String b64Encode (byte[] s_bytes) {
		int i;
		byte[] inBuf = new byte[3];
		byte[] outBuf = new byte[4];
		byte[] result = new byte[((s_bytes.length + 2) / 3) * 4];
		for (i = 1; i <= (s_bytes.length + 2) / 3; i++) {
			if (s_bytes.length < i * 3) {
				inBuf[0] = 0;
				inBuf[1] = 0;
				inBuf[2] = 0;
				System.arraycopy (s_bytes, (i - 1) * 3, inBuf, 0,
					s_bytes.length - (i - 1) * 3
				);
			} else {
				System.arraycopy (s_bytes, (i - 1) * 3, inBuf, 0, 3);
			}
			outBuf[0] = B64TABLE[ (inBuf[0] & 0xFC) >> 2];
			outBuf[1]
			 = B64TABLE[((inBuf[0] & 0x03) << 4) | ((inBuf[1] & 0xF0) >> 4)];
			outBuf[2]
			 = B64TABLE[((inBuf[1] & 0x0F) << 2) | ((inBuf[2] & 0xC0) >> 6)];
			outBuf[3] = B64TABLE[  inBuf[2] & 0x3F];
			System.arraycopy (outBuf, 0, result, (i - 1) * 4, 4);
		}
		if ((s_bytes.length % 3) == 1) {
			result[result.length - 2] = B64TABLE[64];
			result[result.length - 1] = B64TABLE[64];
		} else if ((s_bytes.length % 3) == 2) {
			result[result.length - 1] = B64TABLE[64];
		}
		return new String (result);
	}//b64Encode (...)
	
	/**
	 * decodiert einen Base64-String in ein byte-Array
	 * 
	 * @param s der zu decodierende String
	 * @return die decodierten Daten als byte-Array
	 * @exception IllegalArgumentException wenn der String ungltige Zeichen
	 *            enthlt oder eine ungltige Lnge hat
	 */
	public static byte[] b64Decode (String s) {
		if ((s.length () % 4) != 0) {
			throw (new IllegalArgumentException (
				"b64Decode: illegal String format"
			));
		}
		int i, j;
		byte[] s_bytes = s.getBytes ();
		byte[] inBuf = new byte[4];
		byte[] outBuf = new byte[3];
		byte[] result = new byte[0];
		if (s.length () > 3) {
			result = new byte[((s.length () / 4) - 1) * 3];
		}
		byte[] res = null;
		for (i = 1; i < (s.length () / 4); i++) {
			System.arraycopy (s_bytes, (i - 1) * 4, inBuf, 0, 4);
			for (j = 0; j < 4; j++) {
				if ((inBuf[j] > 64) && (inBuf[j] < 91)) {
					inBuf[j] -= 65;
				} else if ((inBuf[j] > 96) && (inBuf[j] < 123)) {
					inBuf[j] -= 71;
				} else if ((inBuf[j] > 47) && (inBuf[j] < 58)) {
					inBuf[j] += 4;
				} else if (inBuf[j] == 43) {
					inBuf[j] = 62;
				} else if (inBuf[j] == 47) {
					inBuf[j] = 63;
				} else {
					throw (new IllegalArgumentException (
						"b64Decode: illegal String format"
					));
				}
			}
			outBuf[0]
			 = (byte)(((inBuf[0] << 2) & 0xFC) | ((inBuf[1] >> 4) & 0x03));
			outBuf[1]
			 = (byte)(((inBuf[1] << 4) & 0xF0) | ((inBuf[2] >> 2) & 0x0F));
			outBuf[2] = (byte)(((inBuf[2] << 6) & 0xC0) | (inBuf[3] & 0x3F));
			System.arraycopy (outBuf, 0, result, (i - 1) * 3, 3);
		}
		if (s.length () != 0) {
			System.arraycopy (s_bytes, s.length () - 4, inBuf, 0, 4);
			if (inBuf[2] == 61) {
				for (j = 0; j < 2; j++) {
					if ((inBuf[j] > 64) && (inBuf[j] < 91)) {
						inBuf[j] -= 65;
					} else if ((inBuf[j] > 96) && (inBuf[j] < 123)) {
						inBuf[j] -= 71;
					} else if ((inBuf[j] > 47) && (inBuf[j] < 58)) {
						inBuf[j] += 4;
					} else if (inBuf[j] == 43) {
						inBuf[j] = 62;
					} else if (inBuf[j] == 47) {
						inBuf[j] = 63;
					} else {
						throw (new IllegalArgumentException (
							"b64Decode: illegal String format"
						));
					}
				}
				outBuf[0]
				 = (byte)(((inBuf[0] << 2) & 0xFC) | ((inBuf[1] >> 4) & 0x03));
				res = new byte[result.length + 1];
				System.arraycopy (result, 0, res, 0, result.length);
				System.arraycopy (outBuf, 0, res, result.length, 1);
			} else if (inBuf[3] == 61) {
				for (j = 0; j < 3; j++) {
					if ((inBuf[j] > 64) && (inBuf[j] < 91)) {
						inBuf[j] -= 65;
					} else if ((inBuf[j] > 96) && (inBuf[j] < 123)) {
						inBuf[j] -= 71;
					} else if ((inBuf[j] > 47) && (inBuf[j] < 58)) {
						inBuf[j] += 4;
					} else if (inBuf[j] == 43) {
						inBuf[j] = 62;
					} else if (inBuf[j] == 47) {
						inBuf[j] = 63;
					} else {
						throw (new IllegalArgumentException (
							"b64Decode: illegal String format"
						));
					}
				}
				outBuf[0]
				 = (byte)(((inBuf[0] << 2) & 0xFC) | ((inBuf[1] >> 4) & 0x03));
				outBuf[1]
				 = (byte)(((inBuf[1] << 4) & 0xF0) | ((inBuf[2] >> 2) & 0x0F));
				res = new byte[result.length + 2];
				System.arraycopy (result, 0, res, 0, result.length);
				System.arraycopy (outBuf, 0, res, result.length, 2);
			} else {
				for (j = 0; j < 4; j++) {
					if ((inBuf[j] > 64) && (inBuf[j] < 91)) {
						inBuf[j] -= 65;
					} else if ((inBuf[j] > 96) && (inBuf[j] < 123)) {
						inBuf[j] -= 71;
					} else if ((inBuf[j] > 47) && (inBuf[j] < 58)) {
						inBuf[j] += 4;
					} else if (inBuf[j] == 43) {
						inBuf[j] = 62;
					} else if (inBuf[j] == 47) {
						inBuf[j] = 63;
					} else {
						throw (new IllegalArgumentException (
							"b64Decode: illegal String format"
						));
					}
				}
				outBuf[0]
				 = (byte)(((inBuf[0] << 2) & 0xFC) | ((inBuf[1] >> 4) & 0x03));
				outBuf[1]
				 = (byte)(((inBuf[1] << 4) & 0xF0) | ((inBuf[2] >> 2) & 0x0F));
				outBuf[2]
				 = (byte)(((inBuf[2] << 6) & 0xC0) | (inBuf[3] & 0x3F));
				res = new byte[result.length + 3];
				System.arraycopy (result, 0, res, 0, result.length);
				System.arraycopy (outBuf, 0, res, result.length, 3);
			}
		}
		return res;
	}//b64Decode (...)
	
	/**
	 * wandelt ein byte-Array in einen String
	 * 
	 * @param s_bytes die zu codierenden Daten
	 * @return die codierten Daten als String
	 */
	public static String plainEncode (byte[] s_bytes) {
		return new String (s_bytes);
//		return new String (s_bytes, "ISO_8859-15");
	}
	
	/**
	 * wandelt ein String in ein byte-Array
	 * 
	 * @param s der zu bearbeitende String
	 * @return der STring als Byte-Array
	 */
	public static byte[] plainDecode (String s) {
		return s.getBytes ();
//		return s.getBytes ("ISO_8859-15");
	}
}//B64_Hex