package ec3.shared.crypto;

/**
 * die Klasse implementiert den SHA256-Prfsummenlagorithmus
 * 
 * @author Julian Boot
 * @author AM
 */
public class SHA256 {
	
	/**
	 * die H-Konstanten des Algorithmus (Startwerte fr die Prfsumme)
	 */
	private static final int[] H = new int[] {
		0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a,
		0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19
	};
	
	/**
	 * die K-Konastanten des Algorithmus
	 */
	private static final int[] K = {
		0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
		0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
		0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
		0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
		0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
		0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
		0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
		0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
		0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
		0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
		0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
		0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
		0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
		0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
		0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
		0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
	};
	
	/**
	 * berechnet die Prfsumme des angegeben Byte-Arrays
	 * 
	 * @param data die Daten, ber die die Prfsumme zu bilden ist
	 * @return die Prfsumme
	 */
	public static byte[] digest (byte[] data) {
		int[] localH = new int[8];
		int[] W = new int[64];
		byte padded[] = padData(data);
		// intermediate hash result, initialised with constants
		System.arraycopy(SHA256.H, 0, localH, 0, 8);
		int blocks = padded.length / 64;
		for(int i=0; i < blocks; i++) {
			// build message schedule
			for(int j=0; j < 16; j++) {
				W[j] =  bytesToInt(padded, (i * 64) + (j * 4));
			}
			for(int j=16; j < 64; j++) {
				W[j] = sigma1(W[j-2]) + W[j-7] + sigma0(W[j-15]) + W[j-16];
			}
			// do compression
			int a = localH[0];
			int b = localH[1];
			int c = localH[2];
			int d = localH[3];
			int e = localH[4];
			int f = localH[5];
			int g = localH[6];
			int h = localH[7];
			for(int j=0; j < 64; j++) {
				int tmp1 = h + SIGMA1(e) + Ch(e,f,g) + K[j] + W[j];
				int tmp2 = SIGMA0(a) + Maj(a,b,c);
				h = g;
				g = f;
				f = e;
				e = d + tmp1;
				d = c;
				c = b;
				b = a;
				a = tmp1 + tmp2;
			}
			localH[0] += a;
			localH[1] += b;
			localH[2] += c;
			localH[3] += d;
			localH[4] += e;
			localH[5] += f;
			localH[6] += g;
			localH[7] += h;
		}
		// build the result
		byte hash[] = new byte[32];
		for (int i=0; i < 8 ; i ++) {
			encode32Bits(localH[i], hash, i * 4);
		}
		return hash;
	}
	
	/**
	 * ergnzt die Daten, so da die Lnge ein ganzzahliges Vielfaches von
	 * 64 Byte ist
	 * 
	 * @param msg die Orginadaten
	 * @return die ergnzteen Daten
	 */
	private static final byte[] padData(byte[] msg) {
		int size = msg.length;
		int pad = 64 - (size % 64);
		if (pad < 9) {
			// not enough space - add new 64byte block
			pad += 64;
			
		}
		byte[] result = new byte[size + pad];
		System.arraycopy(msg, 0, result, 0, size);
		result[size] |= (0x1 << 7); // bit after message is 1
		encode64Bits(size * 8, result, (size + pad - 8));
		return result;
	}
	
	/**
	 * setzt 4 bytes zu einem int zusammen
	 * 
	 * @param buf die bytes
	 * @param offset ab wo soll gearbeitet werden
	 * @return der zusammengesetzte int
	 */
	private static final int bytesToInt(byte[] buf, int offset) {
		int result = 0;
		int radix = 0;
		int i = 4;
		do {
			int temp = buf[offset + (--i)];
		    result |= (temp & 0xFF) << radix;
		    radix += 8;
		} while (i > 0);
		return result;
	}
	
	/**
	 * die Funktion Ch des SHA256-Algorithmus
	 * 
	 * @param x der x-Parameter
	 * @param y der y-Parameter
	 * @param z der z-Parameter
	 * @return das Ergebnis
	 */
	private static final int Ch(int x, int y, int z) {
		return (x & y) ^ (~x & z);
	}
	
	/**
	 * die Funktion Maj des SHA256-Algorithmus
	 * 
	 * @param x der x-Parameter
	 * @param y der y-Parameter
	 * @param z der z-Parameter
	 * @return das Ergebnis
	 */
	private static final int Maj(int x, int y, int z) {
		return (x & y) ^ (x & z) ^ (y & z);
	}

	/**
	 * die Funktion SIGMA0 des SHA256-Algorithmus
	 * 
	 * @param x der x-Parameter
	 * @return das Ergebnis
	 */
	private static final int SIGMA0(int x) {
		return S(2,x) ^ S(13, x) ^ S(22, x);
	}

	/**
	 *die Funktion SIGMA1 des SHA256-Algorithmus
	 * 
	 * @param x der x-Parameter
	 * @return das Ergebnis
	 */
	private static final int SIGMA1(int x) {
		return S(6, x) ^ S(11, x) ^ S(25, x);
	}

	/**
	 * die Funktion sigma0 des SHA256-Algorithmus
	 * 
	 * @param x der x-Parameter
	 * @return das Ergebnis
	 */
	private static final int sigma0(int x) {
		return S(7, x) ^ S(18, x) ^ R(3, x);
	}

	/**
	 * die Funktion sigma1 des SHA256-Algorithmus
	 * 
	 * @param x der x-Parameter
	 * @return das Ergebnis
	 */
	private static final int sigma1(int x) {
		return S(17, x) ^ S(19, x) ^ R(10, x);
	}
	
	/**
	 * rotiert eine Zahl um n Bits nach rechts
	 * 
	 * (This is called 'S' in the spec, rather than 'R')
	 * 
	 * @param n die Bitzahl, um die rotiert wird
	 * @param x die zu rotierende Zahl
	 * @return die rotierte Zahl
	 */
	private static final int S(int n, int x) {
		return (x >>> n) | (x << (-n));
	}
	
	/**
	 * verschiebt eine Zahl um n Bits nach rechts
	 * 
	 * (This is called 'R' in the spec, rather than 'S')
	 * 
	 * @param n die Bitzahl, um die verschoben wird
	 * @param x die zu verschiebende Zahl
	 * @return die verschobene Zahl
	 */
	private static final int R(int n, int x) {
		
		return (x >>> n);
		
	}
	
	/**
	 * Zerlegt einen int in 4 Bytes
	 * 
	 * @param value der zu zerlegende int
	 * @param buf der Zielpuffer fr die Bytes
	 * @param offset der Offset, ab dem geschrieben werden soll
	 */
	private static final void encode32Bits(int value, byte[] buf, int offset) {
		int i = 4;
		int mask = 0xFF;
		int radix = 0x00;
		do {
			buf[(offset + (--i))] = (byte) ((value & mask) >>> radix);
			mask <<= 8;
			radix += 8;
		} while (i > 0);
	}
	
	/**
	 * Zerlegt einen long in 8 Bytes
	 * 
	 * @param value der zu zerlegende long
	 * @param buf der Zielpuffer fr die Bytes
	 * @param offset der Offset, ab dem geschrieben werden soll
	 */
	private static final void encode64Bits(long value, byte[] buf, int offset) {
		int i = 8;
		int mask = 0xFF;
		int radix = 0x00;
		do {
			buf[(offset + (--i))] = (byte) ((value & mask) >>> radix);
			mask <<= 8;
			radix += 8;
		} while (i > 0);
	}
}